/*****************************************************************************
* $Id: atoms.c,v 1.10 2004/05/08 23:38:32 playfair Exp $
 *****************************************************************************
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111,
 * USA.
 *****************************************************************************/

#include <fcntl.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>

#include "config.h"

#include "hymn.h"
#include "hymn_private.h"

#include "endianutils.h"
#include "atoms.h"

#include "aes.h"
#include "md5.h"

char* get_homedir( void );
int get_user_key( char *psz_homedir, uint32_t i_user, uint32_t i_key,
                  uint32_t *p_user_key );

void decrypt(uint8_t *input, uint8_t *output, uint32_t len,
             uint8_t *key, uint8_t *iv) {
    aes_context ctx;
    aes_set_key(&ctx, key, 128);
    aes_cbc_decrypt(&ctx, input, output, len, iv);
}

bool is_container(uint32_t atom_id) {
    switch (atom_id) {
    case ATOM_ALB:
    case ATOM_ART:
    case ATOM_CLIP:
    case ATOM_CMT:
    case ATOM_COVR:
    case ATOM_CPIL:
    case ATOM_DAY:
    case ATOM_DINF:
    case ATOM_DISK:
    case ATOM_DRMS:
    case ATOM_EDTS:
    case ATOM_ILST:
    case ATOM_MATT:
    case ATOM_MDIA:
    case ATOM_MINF:
    case ATOM_MOOV:
    case ATOM_NAM:
    case ATOM_SCHI:
    case ATOM_SINF:
    case ATOM_STBL:
    case ATOM_TMPO:
    case ATOM_TOO:
    case ATOM_TRAK:
    case ATOM_TRKN:
    case ATOM_UDTA:
    case ATOM_WRT:
        return true;
    default:
        return false;
    }
}

void print_atom(hymn_callback_t *hcb, hymn_t *ctx,
                uint32_t atom_id, uint32_t len, uint32_t level) {
    /* only print on first loop */
    if (!ctx->iter) {
        int i;
        char indent[32];
        char atom_name[5];
        bytes_from_uint32(atom_id, atom_name);
        atom_name[4] = 0;

        for (i=0; i<(2*level); ++i) {
            indent[i] = ' ';
        }
        indent[i] = 0;
        hcb->logmsg(hcb->user_data, hcb->log_lvl, LOG_TRACE,
                    "%s%s: %d bytes.", indent, atom_name, len+8);
    }
}

int write_block(hymn_callback_t *hcb, hymn_t *ctx,
                void *data, uint32_t size) {
    if (ctx->iter && !ctx->write_off) {
        fwrite(data, 1, size, ctx->outfile);
    }
    return SUCCESS;
}

int write_header(hymn_callback_t *hcb, hymn_t *ctx,
                 uint32_t len, uint32_t atom_id) {
    if (ctx->iter && !ctx->write_off) {
        int err;
        /* If our atom is in the hierarchy containing the sinf atom,
         * we need to subtrace the size of the sinf atom from the size
         * of the containing atom.
         */
        switch(atom_id) {
        case ATOM_DRMS:
            atom_id = ATOM_MP4A;
            /* no break here */
        case ATOM_STSD:
        case ATOM_STBL:
        case ATOM_MINF:
        case ATOM_MDIA:
        case ATOM_TRAK:
        case ATOM_MOOV:
            len -= ctx->sinf_size;
            /* no break here */
        default:
            len += 8;
            uint8_t scratch[4];
            bytes_from_uint32_be(len, scratch);
            err = write_block( hcb, ctx, scratch, sizeof(scratch) );
            if (err) return err;
            bytes_from_uint32(atom_id, scratch);
            err = write_block( hcb, ctx, scratch, sizeof(scratch) );
            if (err) return err;
            break;
        }
    }
    return SUCCESS;
}

int write_atom(hymn_callback_t *hcb, hymn_t *ctx,
               uint8_t **fp, uint32_t size) {
    return write_block(hcb, ctx, (*fp)-8, size+8);
}

int parse_iviv_atom(hymn_callback_t *hcb, hymn_t *ctx,
                    uint8_t **fp, uint32_t size, uint32_t level) {
    ctx->drms_iviv = malloc(size);
    memcpy(ctx->drms_iviv, *fp, size);
    ctx->drms_iviv_size = size;
    (*fp) += size;
    return SUCCESS;
}

int parse_name_atom(hymn_callback_t *hcb, hymn_t *ctx,
                    uint8_t **fp, uint32_t size, uint32_t level) {
    ctx->drms_name = malloc(size);
    memcpy(ctx->drms_name, *fp, size);
    ctx->drms_name_size = strlen(ctx->drms_name);
    (*fp) += size;
    return SUCCESS;
}

int parse_priv_atom(hymn_callback_t *hcb, hymn_t *ctx,
                    uint8_t **fp, uint32_t size, uint32_t level) {
    ctx->drms_priv = malloc(size);
    memcpy(ctx->drms_priv, *fp, size);
    ctx->drms_priv_size = size;
    (*fp) += size;
    return SUCCESS;
}

int parse_drms_atom(hymn_callback_t *hcb, hymn_t *ctx,
                    uint8_t **fp, uint32_t size, uint32_t level) {
    int err;
    uint32_t atom_id, len;

    err = write_block( hcb, ctx, *fp, (6+2+8+2+2+4+4) ); /* see below */
    if (err) return err;

    (*fp) += 6; /* reserved */
    (*fp) += 2; /* data reference index */
    (*fp) += 8; /* version, revision, vendor */
    (*fp) += 2; /* channel count */
    (*fp) += 2; /* sample size */
    (*fp) += 4; /* compression id, packet size */
    (*fp) += 4; /* sample rate */

    /* esds */
    err = parse_atom(hcb, ctx, fp, size, level+1);
    if (err) return err;

    /* sinf */
    err = parse_atom(hcb, ctx, fp, size, level+1);
    if (err) return err;

    return SUCCESS;
}

int parse_stsd_atom(hymn_callback_t *hcb, hymn_t *ctx,
                    uint8_t **fp, uint32_t size, uint32_t level) {
    int err;
    uint32_t entry_count, len, atom_id, i;

    err = write_block( hcb, ctx, *fp, (1+3+4) );
    if (err) return err;

    (*fp) += 1; /* version */
    (*fp) += 3; /* flags */
    entry_count = uint32_from_bytes_be(*fp);
    (*fp) += 4;

    /* deal w/ entry count */
    for (i=0; i<entry_count; ++i) {
        uint32_t len = uint32_from_bytes_be(*fp) - 8;
        (*fp) += 4;
        uint32_t atom_id = uint32_from_bytes(*fp);
        (*fp) += 4;

        switch(atom_id) {
        case ATOM_DRMS:
            print_atom(hcb, ctx, atom_id, len, level+1);
            err = write_header(hcb, ctx, len, atom_id);
            if (err) return err;
            err = parse_drms_atom(hcb, ctx, fp, len, level+1);
            if (err) return err;
            break;
        default:
            /* copy atom to outfile */
            err = write_atom(hcb, ctx, fp, len);
            if (err) return err;
            (*fp) += len;
            break;
        }
    }

    return SUCCESS;
}

int parse_stsz_atom(hymn_callback_t *hcb, hymn_t *ctx,
                    uint8_t **fp, uint32_t size, uint32_t level) {
    int err;
    uint32_t i;

    /* copy atom to outfile */
    err = write_atom(hcb, ctx, fp, size);
    if (err) return err;

    /* get the sample table sizes */
    /* first 12 bytes is some sort of header */
    (*fp) += 12;
    size -= 12;
    ctx->num_sample_sizes = size / 4;
    ctx->sample_table_sizes = malloc(size);
    for (i=0; i<ctx->num_sample_sizes; ++i) {
        ctx->sample_table_sizes[i] = uint32_from_bytes_be(*fp);
        (*fp) += 4;
    }
    return SUCCESS;
}

int parse_stco_atom(hymn_callback_t *hcb, hymn_t *ctx,
                    uint8_t **fp, uint32_t size, uint32_t level) {
    /* The first time through, just skip over the stco atom.  The
     * second time through, we need to subtract the sinf atom size
     * from each sample offset.
     */
    if (!ctx->iter) {
        uint32_t i;

        /* get the sample table offsets */
        /* first 8 bytes is some sort of header */
        (*fp) += 8;
        size -= 8;
        ctx->num_sample_offsets = size / 4;
        ctx->sample_table_offsets = malloc(size);
        for (i=0; i<ctx->num_sample_offsets; ++i) {
            ctx->sample_table_offsets[i] = uint32_from_bytes_be(*fp);
            (*fp) += 4;
        }
    } else {
        int err;
        uint32_t i;
        uint8_t block[4];

        err = write_block(hcb, ctx, (*fp)-8, 16);
        if (err) return err;
        for (i=0; i<ctx->num_sample_offsets; ++i) {
            ctx->sample_table_offsets[i] -= (ctx->sinf_size);
            bytes_from_uint32_be(ctx->sample_table_offsets[i], block);
            hcb->logmsg(hcb->user_data, hcb->log_lvl, LOG_SPEW,
                        "sample_table_offsets[%d]: %d,%x%x%x%x",
                        i, ctx->sample_table_offsets[i],
                        block[0], block[1], block[2], block[3]);
            err = write_block(hcb, ctx, block, 4);
            if (err) return err;
        }
        (*fp) += size;
    }
    return SUCCESS;
}

int parse_mdat_atom(hymn_callback_t *hcb, hymn_t *ctx,
                    uint8_t **fp, uint32_t size, uint32_t level) {
    /* The first time through, just skip over the mdat atom.  The
     * second time through, decrypt and write it out to the output
     * file.
     */
    if (!ctx->iter) {
        (*fp) += size;
    } else {
        int i, err;
        char *homedir;
        uint8_t user_key[16];
        uint8_t name_iviv_hash[16], key[16], iv[16], block[4096];

        /* ensure the IPOD environment variable is set if necessary */
        if (! set_keyenv(hcb, ctx->drms_user, ctx->drms_key) ) {
            return hcb->errmsg(hcb->user_data, ERR_NOKEY);
        }

        homedir = get_homedir();
        if ( get_user_key(homedir, ctx->drms_user, ctx->drms_key,
                          (uint32_t*)user_key) ) {
            return hcb->errmsg(hcb->user_data, ERR_GET_USER_KEY);
        }
        free(homedir);
        /* drms code is assuming the opposite endian-ness */
        REVERSE((uint32_t*)user_key, 4);
        hcb->logmsg(hcb->user_data, hcb->log_lvl, LOG_DEBUG,
                    "Got user key: %x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x",
                    user_key[0], user_key[1], user_key[2], user_key[3],
                    user_key[4], user_key[5], user_key[6], user_key[7],
                    user_key[8], user_key[9], user_key[10], user_key[11],
                    user_key[12], user_key[13], user_key[14], user_key[15]);

        md5_context md5;

        hcb->logmsg(hcb->user_data, hcb->log_lvl, LOG_TRACE,
                    "Writing mdat header.");
        err = write_header(hcb, ctx, size, ATOM_MDAT);
        if (err) return err;

        /* hash the name and iviv */
        hcb->logmsg(hcb->user_data, hcb->log_lvl, LOG_TRACE,
                    "Hashing name and iviv.");
        md5_starts(&md5);
        md5_update( &md5, ctx->drms_name, strlen(ctx->drms_name) );
        md5_update(&md5, ctx->drms_iviv, ctx->drms_iviv_size);
        md5_finish(&md5, name_iviv_hash);

        hcb->logmsg(hcb->user_data, hcb->log_lvl, LOG_TRACE,
                    "Decrypting priv.");
        decrypt(ctx->drms_priv, ctx->drms_priv, ctx->drms_priv_size,
                user_key, name_iviv_hash);

        if (ctx->drms_priv[0] != 'i' ||
            ctx->drms_priv[1] != 't' ||
            ctx->drms_priv[2] != 'u' ||
            ctx->drms_priv[3] != 'n') {
            return hcb->errmsg(hcb->user_data, ERR_KEY_DECRYPT);
        } else {
            memcpy(key, ctx->drms_priv+24, 16);
            memcpy(iv, ctx->drms_priv+48, 16);

            hcb->logmsg(hcb->user_data, hcb->log_lvl, LOG_TRACE,
                        "Decrypting mdat.");
            for (i=0; i<ctx->num_sample_sizes; ++i) {
                hcb->logmsg(hcb->user_data, hcb->log_lvl, LOG_SPEW,
                            "sample_table[%d]: %d bytes.",
                            i, ctx->sample_table_sizes[i]);
                decrypt(*fp, block, ctx->sample_table_sizes[i], key, iv);
                err = write_block(hcb, ctx, block, ctx->sample_table_sizes[i]);
                if (err) return err;
                (*fp) += ctx->sample_table_sizes[i];
            }
        }
    }
    return SUCCESS;
}

int parse_atom(hymn_callback_t *hcb, hymn_t *ctx,
               uint8_t **fp, uint32_t size, uint32_t level) {
    int err;
    uint32_t len, atom_id;

    len = uint32_from_bytes_be(*fp) - 8;
    (*fp) += 4;
    atom_id = uint32_from_bytes(*fp);
    (*fp) += 4;

    print_atom(hcb, ctx, atom_id, len, level);

    switch(atom_id) {
    case ATOM_STSD:
        err = write_header(hcb, ctx, len, atom_id);
        if (err) return err;
        err = parse_stsd_atom(hcb, ctx, fp, len, level);
        if (err) return err;
        break;
    case ATOM_USER:
        ctx->drms_user = uint32_from_bytes_be(*fp);
        (*fp) += 4;
        break;
    case ATOM_KEY:
        ctx->drms_key = uint32_from_bytes_be(*fp);
        (*fp) += 4;
        break;
    case ATOM_IVIV:
        err = parse_iviv_atom(hcb, ctx, fp, len, level+1);
        if (err) return err;
        break;
    case ATOM_NAME:
        err = parse_name_atom(hcb, ctx, fp, len, level+1);
        if (err) return err;
        break;
    case ATOM_PRIV:
        err = parse_priv_atom(hcb, ctx, fp, len, level+1);
        if (err) return err;
        break;
    case ATOM_STSZ:
        err = parse_stsz_atom(hcb, ctx, fp, len, level+1);
        if (err) return err;
        break;
    case ATOM_MDAT:
        err = parse_mdat_atom(hcb, ctx, fp, len, level+1);
        if (err) return err;
        break;
    case ATOM_STCO:
        err = parse_stco_atom(hcb, ctx, fp, len, level+1);
        if (err) return err;
        break;
    case ATOM_SINF:
        ctx->write_off = true;
        ctx->sinf_size = len + 8;
        err = parse_atoms(hcb, ctx, fp, len, level+1);
        if (err) return err;
        ctx->write_off = false;
        break;
    default:
        if ( is_container(atom_id) ) {
            /* copy header to outfile */
            err = write_header(hcb, ctx, len, atom_id);
            if (err) return err;
            err = parse_atoms(hcb, ctx, fp, len, level+1);
            if (err) return err;
        } else {
            /* copy atom to outfile */
            err = write_atom(hcb, ctx, fp, len);
            if (err) return err;
            (*fp) += len;
        }
        break;
    }

    return SUCCESS;
}

int parse_atoms(hymn_callback_t *hcb, hymn_t *ctx,
                uint8_t **fp, uint32_t len, uint32_t level) {
    int err;
    uint8_t *eob = (*fp) + len;
    while ( (*fp) < eob ) {
        err = parse_atom(hcb, ctx, fp, len, level);
        if (err) return err;
    }

    return SUCCESS;
}

/* EOF */
